#include "MeshLoader.h"
#include <set>
#include <list>
#include <unordered_set>
#include <array>
#include <algorithm>
#include <ranges>
#include<string>
#include <fstream>

MeshLoader::~MeshLoader() //!!! У деструктора указан =default. Зачем тут это определение?
{
}

std::vector<Node>& MeshLoader::GetNodes() {
	return nodes;
}
Node MeshLoader::get_node(rsize_t id) {
    return nodes.at(id);
}


std::vector<FiniteElement>& MeshLoader::GetFE() {
	return finiteElements;
}

std::vector<BoundaryFiniteElement>& MeshLoader::GetBFE() {
	return boundaryFiniteElements;
}

void MeshLoader::Print_Data()
{
    std::cout << &GetNodes() << std::endl
        << &GetFE() << std::endl
        << &GetBFE() << std::endl;
}

//!!! ИСПРАВЛЕНО Контейнеры-результаты должны возвращать ID, а не сами узлы и КЭ

std::vector<int> MeshLoader::findFE(size_t ID1, size_t ID2, size_t ID3) {
    std::vector<int> ans;
    auto cur_iter = finiteElements.begin();
    auto last_iter = finiteElements.end();

    while (cur_iter != last_iter) {
        cur_iter = std::find_if(cur_iter, last_iter, [ID1, ID2, ID3](FiniteElement& cur_elem) { //!!!cur_elem изменяется? Где const?
            return std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID1) != cur_elem.materials.end()
                && std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID2) != cur_elem.materials.end()
                && std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID3) != cur_elem.materials.end();
            });

        if (cur_iter != last_iter) {
            ans.push_back(cur_iter->ID);
            ++cur_iter;
        }
    }

    return ans;

}


std::vector<FiniteElement> MeshLoader::findEdge(size_t ID1, size_t ID2) {
    std::vector<FiniteElement> ans;
    auto cur_iter = finiteElements.begin();
    auto last_iter = finiteElements.end();

    while (cur_iter != last_iter) {
        cur_iter = std::find_if(cur_iter, last_iter, [ID1, ID2](FiniteElement& cur_elem) {//!!!cur_elem изменяется? Где const?
            return std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID1) != cur_elem.materials.end()
                && std::find(cur_elem.materials.begin(), cur_elem.materials.end(), ID2) != cur_elem.materials.end();
            });

        if (cur_iter != last_iter) {
            ans.push_back(*cur_iter);
            ++cur_iter;
        }
    }

    return ans;

}

std::vector<Node> MeshLoader::vertex_nodes_by_surface_ID(size_t s_id) {

        std::vector<Node> answer;
        auto current_iter = boundaryFiniteElements.begin();
        while (current_iter != boundaryFiniteElements.end()) {
            current_iter = std::find_if(current_iter, boundaryFiniteElements.end(), [s_id](const BoundaryFiniteElement& s) { return s.ID_surface == s_id; });
            if (current_iter != boundaryFiniteElements.end())
            {
                for (const auto& elem : current_iter->node_boundary) {
                    
                    answer.insert(Get_Node_by_ID(s_id)); //!!! (Не исправлено) Где гарантия, что не будет дубликатов? Разные КЭ могут содержать одни и те же узлы.
                                                         //!!! ПРЕЖДЕ ЧЕМ ОТПРАВЛЯТЬ НА ПРОВЕРКУ КОМПИЛИРУЙТЕ ПРОГРАММУ!
                                                         //!!! У вектора нет insert с одним параметром.
                }
                ++current_iter;
            }
        }
        return answer;
    }


std::vector<Node> MeshLoader::getMaterialID(size_t ID) {
    std::list<Node> result;
    auto element_by_material = nodes.begin();
    auto last_element = nodes.end();
    while (element_by_material != last_element) {
        
        element_by_material = std::find_if(element_by_material, last_element, [ID](const Node& e) { return e.ID_node == ID; });
        
        if (element_by_material != last_element) {
            result.push_back(*element_by_material);
            ++element_by_material;
        }
    }
    return std::vector<Node>(result.begin(), result.end());
}

std::vector<std::set<size_t>>  MeshLoader::GetNodesNeighbors()
{
    std::vector<std::set<size_t>> result(nodes.size());
    for (const auto& it_FiniteElement : boundaryFiniteElements)
    {
        for (auto& it_nodes : it_FiniteElement.node_boundary)
        {
            result[it_nodes - 1].insert(it_FiniteElement.node_boundary.begin(), it_FiniteElement.node_boundary.end());
            result[it_nodes - 1].erase(it_nodes);
        }
    }

    return result;
}

std::vector<BoundaryFiniteElement> MeshLoader::getBoundaryFiniteID(size_t ID) {
    std::list<BoundaryFiniteElement> result;
    auto surface_by_ID = boundaryFiniteElements.begin();
    auto last_element = boundaryFiniteElements.end();
    while (surface_by_ID != last_element) {
        surface_by_ID = std::find_if(surface_by_ID, last_element, [ID](const BoundaryFiniteElement& s) { return s.ID_surface == ID; });
        if (surface_by_ID != last_element) {
            result.push_back(*surface_by_ID);
            ++surface_by_ID;
        }
    }
    return std::vector<BoundaryFiniteElement>(result.begin(), result.end());
}

Node MeshLoader::Get_Node_by_ID(size_t n_id){ //!!! Лишнее копирование при возврате
    return nodes.at(n_id);
}

void MeshLoader::InsertNewNodes(){
    std::unordered_set<Edge, EdgeHash> edge_container;
    for (auto& elem : boundaryFiniteElements) 
    {
        std::vector<int> current_iter(elem.ID_boundary);
        for (auto i = current_iter.begin(); i != current_iter.end(); i++)
        {
            for (auto j = i + 1; j != current_iter.end(); j++)
            {
                Edge current_edge(*i, *j);
                auto search_edge = edge_container.find(current_edge);

                if (search_edge == edge_container.end()) {
                    Node left_node = Get_Node_by_ID(*i); //!!! Копрование узла на каждой итерации.
                    Node right_node = Get_Node_by_ID(*j); //!!! Копрование узла на каждой итерации.

                    //!!! Форматируйте код. Он не должнен быть шире 100 символов
                    std::vector<float> x = { (left_node.coord[0] + right_node.coord[0]) / 2, (left_node.coord[0] + right_node.coord[0]) / 2, (left_node.coord[0] + right_node.coord[0]) / 2 };
                    Node center_node(nodes.size() + 1, x, false); //!!! Копирование x в узел. Либо используйте move (с нужной версией конструктора),
                                                                  //!!! либо создавайте узел и редактируйте его поле, а не заренее вектор создавайте.
                    current_edge.self_id = nodes.size() + 1;
                    edge_container.insert(current_edge);
                    nodes.push_back(center_node);
                    elem.node_boundary.emplace_back(current_edge.self_id);
                }
                else{
                    elem.node_boundary.emplace_back(search_edge->self_id);
                }

            }
        }
    }

    for (auto& it : boundaryFiniteElements)
    {
        std::vector<size_t> current_iter(it.node_boundary);
        for (auto i = current_iter.begin(); i != current_iter.end() - 1; i++)
        {
            for (auto j = i + 1; j != current_iter.end(); j++)
            {
                Edge search_edge(*i, *j);
                auto search_point = edge_container.find(search_edge);
                if (search_point != edge_container.end())
                {
                    it.node_boundary.emplace_back(search_point->self_id);
                }
            }
        }
    }


}

bool MeshLoader::ratio(const FiniteElement& elem) {
    if (elem.materials.size() < 4) return false;
    Node A = get_node(elem.materials[0]);
    Node B = get_node(elem.materials[1]);
    Node C = get_node(elem.materials[2]);
    Node D = get_node(elem.materials[3]);
    auto len = [](Node x, Node y) {
        return sqrt((x.coord[0] - y.coord[0]) * (x.coord[0] - y.coord[0]) 
                  + (x.coord[1] - y.coord[1]) * (x.coord[1] - y.coord[1]) 
                  + (x.coord[2] - y.coord[2]) * (x.coord[2] - y.coord[2]));
    };
    float A1 = std::max(std::max(std::max(len(A,B), len(A, C)), std::max(len(A, D), len(B, C))), std::max(len(B, D), len(C, D)));
    float A2 = (0.5 * (3 * A1)) * A1 + (sqrt(3) / (4 * (A1 * A1)));
    return A2 / A1>1;
}

void MeshLoader::sort_range(const std::string& filepath) {

    std::vector<FiniteElement> elements = GetFE();
    if (elements.empty())
        return;

    // !!! Где сортировка по данному притерию?

    std::ofstream file(filepath + ".txt");


    std::ranges::copy(elements | std::ranges::views::filter([&](const FiniteElement& ABCD) { //!!! По заданию нужно взять первые 10, а не фильтровать.
        return ratio(ABCD);
        }
    ) | std::ranges::views::transform(
            [](const FiniteElement& ABCD) {
                return ABCD.materials; //!!! Выводит по заданию нужно список ID вершин. Если это он, то при чем тут материалы?
            }), std::ostream_iterator<std::vector<size_t>>(file)); //!!! Где перегрузка << для std::vector<size_t>?

}
